unit WPRuler;

interface

uses
  Windows, Messages, Forms, SysUtils, Classes, Graphics, Controls, Dialogs,
  ExtCtrls, Printers, WPTabEdit, Dsgnintf, Menus, WinSpool;

const
  MaxTabs = 50; //Specifies the maximum number of tabs. Can be changed to any
                //BYTE value (i.e. not more than 255)

//{$DEFINE UKVERSION} //Enable this define to make hint and popup menu
                      //spellings UK English (Centre instead of Center)

type
  TSliderEvent = procedure(Sender: TObject; SliderName: string; Margin: Single)
    of object;
  TTabEvent = procedure(Sender: TObject; Index: Byte) of object;
  TSliderStyle = (ssUp, ssDown, ssMulti);
  TTabStyle = (tsLeftAlign, tsCenterAlign, tsRightAlign);
  TUnits = (unInches, unCentimeters);
  TDummy = class(TPersistent);
  TTabRec = record
    Pos: Single;
    Style: TTabStyle;
  end;
  TTabsEditor = class(TPropertyEditor)
    TabForm: TWPTabForm;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure Edit; override;
  end;
  TFolderEditor = class(TPropertyEditor)
    OpenDialog: TOpenDialog;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
    procedure Edit; override;
  end;
  TWPSlider = class(TGraphicControl)
  private
    FSliderStyle: TSliderStyle;
    FFillColor: TColor;
    procedure SetFillColor(Value: TColor);
    procedure SetSliderStyle(Value: TSliderStyle);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property FillColor: TColor read FFillColor write SetFillColor;
    property Hint;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property ShowHint;
    property SliderStyle: TSliderStyle read FSliderStyle write SetSliderStyle;
  end;
  TWPTab = class(TGraphicControl)
  private
    FTabPos: Single;
    FTabStyle: TTabStyle;
    FColor: TColor;
    procedure SetColor(Value: TColor);
    procedure SetTabStyle(Value: TTabStyle);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Color: TColor read FColor write SetColor;
    property Hint;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property TabPos: Single read FTabPos write FTabPos;
    property TabStyle: TTabStyle read FTabStyle write SetTabStyle;
  end;
  TWPRuler = class(TCustomControl)
  private
    { Private declarations }
    FLMargin, FRMargin, FIMargin, FMulti: TWPSlider;
    FLeftMargin, FRightMargin, FIndent, FPageWidth, PrinterWidth: Single;
    FLeftInset: Word;
    FNumber: Byte;
    FUnits: TUnits;
    FOnSliderMoved: TSliderEvent;
    FOnTabAdded, FOnTabMoved: TTabEvent;
    FShowLM, FShowRM, FShowID, FShowMU: Boolean;
    FTabObjects: array[0..MaxTabs - 1] of TWPTab;
    FScreenRes: Word;
    FNumTabs: Byte;
    FColor, FTabColor: TColor;
    FProjectFolder: string;
    FSort, Dragging: Boolean;
    Diff: Integer;
    Timer: TTimer;
    Popup: TPopupMenu;
    PopupItems: array[0..2] of TMenuItem;
    procedure AddOneTab(X: Single; Style: TTabStyle);
    procedure RemoveOneTab(Index: Byte);
    procedure DoTabs(Value: Boolean);
    function GetTabs: TDummy;
    function GetTabPixels(Index: Byte): Word;
    function GetTabStyles(Index: Byte): TTabStyle;
    function GetTabValues(Index: Byte): Single;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;
    procedure SetColor(Value: TColor);
    procedure SetIndent(Value: Single);
    procedure SetLeftMargin(Value: Single);
    procedure SetLeftInset(Value: Word);
    procedure SetRightMargin(Value: Single);
    procedure SetShowLM(Value: Boolean);
    procedure SetShowRM(Value: Boolean);
    procedure SetShowID(Value: Boolean);
    procedure SetShowMU(Value: Boolean);
    procedure SetSort(Value: Boolean);
    procedure SetTabColor(Value: TColor);
    procedure SetTabs(Value: TDummy);
    procedure SetTabStyles(Index: Byte; Value: TTabStyle);
    procedure SetTabValues(Index: Byte; Value: Single);
    procedure SetUnits(Value: TUnits);
    procedure SliderMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SliderMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SliderMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure TabMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TabMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TabMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure PopupClick(Sender: TObject);
    procedure SortTabs;
    procedure TimerProc(Sender: TObject);
  protected
    { Protected declarations }
    property ScreenRes: Word read FScreenRes;
    property TabsChanged: Boolean write DoTabs;
    procedure Loaded; override;
    procedure Paint; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddTab(Position: Single; Style: TTabStyle): Boolean;
    function RemoveTab(Index: Byte): Boolean;
    procedure ClearAllTabs;
    property Canvas;
    property NumTabs: Byte read FNumTabs;
    property PageWidth: Single read FPageWidth;
    property Sort: Boolean read FSort write SetSort;
    property TabPixels[Index: Byte]: Word read GetTabPixels;
    property TabStyles[Index: Byte]: TTabStyle read GetTabStyles write SetTabStyles;
    property TabValues[Index: Byte]: Single read GetTabValues write SetTabValues;
  published
    { Published declarations }
    property Align;
    property Color: TColor read FColor write SetColor;
    property Font;
    property Hint;
    property Units: TUnits read FUnits write SetUnits default unCentimeters;
    property Indent: Single read FIndent write SetIndent;
    property LeftMargin: Single read FLeftMargin write SetLeftMargin;
    property LeftInset: Word read FLeftInset write SetLeftInset default 10;
    property ParentFont;
    property ParentShowHint;
    property ProjectFolder: string read FProjectFolder write FProjectFolder;
    property RightMargin: Single read FRightMargin write SetRightMargin;
    property ShowHint;
    property ShowLMSlider: Boolean read FShowLM write SetShowLM default True;
    property ShowRMSlider: Boolean read FShowRM write SetShowRM default True;
    property ShowIDSlider: Boolean read FShowID write SetShowID default True;
    property ShowLMandIDSlider: Boolean read FShowMU write SetShowMU default True;
    property Tabs: TDummy read GetTabs write SetTabs; //Dummy property
    property TabColor: TColor read FTabColor write SetTabColor;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnSliderMoved: TSliderEvent read FOnSliderMoved write FOnSliderMoved;
    property OnTabAdded: TTabEvent read FOnTabAdded write FOnTabAdded;
    property OnTabMoved: TTabEvent read FOnTabMoved write FOnTabMoved;
  end;

procedure Register;

implementation

const
  DatFile = 'WPRuler.dat';

procedure Register;
begin
  RegisterComponents('EWC Comps', [TWPRuler]);
  RegisterPropertyEditor(TypeInfo(TDummy), TWPRuler, 'Tabs',
    TTabsEditor);
  RegisterPropertyEditor(TypeInfo(string), TWPRuler, 'ProjectFolder',
    TFolderEditor);
end;

//Tabs Property Editor
//====================
//Get property attributes

function TTabsEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

//Display text in Object Inspector

function TTabsEditor.GetValue: string;
begin
  Result := '(TTabValues)';
end;

//Display tabs property editor dialog

procedure TTabsEditor.Edit;
var
  N: Byte;
  C: TWPRuler;
  Value: Single;
  Change: Boolean;
  TS: Char;
const
  Styles: array[0..2] of Char = ('L', 'C', 'R');
begin
  Change := False;
  TabForm := TWPTabForm.Create(nil);
  C := TWPRuler(GetComponent(0));
  with TabForm do
  begin
    if C.Units = unInches then UnitLabel.Caption := 'Inches'
    else UnitLabel.Caption := 'Centimeters';
    for N := 0 to MaxTabs - 1 do
      with C do
      begin
        TabGrid.Cells[1, N + 1] := FloatToStrF(TabValues[N], ffGeneral, 5, 0);
        TabGrid.Cells[2, N + 1] := Styles[Byte(TabStyles[N])];
      end;
    ShowModal;
    if ModalResult = mrOK then
      with C do
      begin
        Sort := False;
        for N := 0 to MaxTabs - 1 do
        begin
          if TabGrid.Cells[1, N + 1] = '' then Value := 0
          else Value := StrToFloat(TabGrid.Cells[1, N + 1]);
          if TabValues[N] <> Value then Change := True;
          TabValues[N] := Value;
          if TabGrid.Cells[2, N + 1] = '' then TS := 'L'
          else TS := TabGrid.Cells[2, N + 1][1];
          if TS <> Styles[Byte(TabStyles[N])] then Change := True;
          if TS = 'L' then TabStyles[N] := tsLeftAlign
          else if TS = 'C' then TabStyles[N] := tsCenterAlign
          else TabStyles[N] := tsRightAlign;
        end;
        Sort := True;
      end;
  end;
  TabForm.Release;
  if Change then C.TabsChanged := True;
end;

//ProjectFolder Property Editor
//=============================
//Get property attributes

function TFolderEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

//Display text in Object Inspector

function TFolderEditor.GetValue: string;
begin
  Result := GetStrValue;
end;

//Change the text using the Object Inspector

procedure TFolderEditor.SetValue(const Value: string);
var
  C: TWPRuler;
begin
  SetStrValue(Value);
  C := TWPRuler(GetComponent(0));
  C.FProjectFolder := Value;
end;

//Get the project folder path

procedure TFolderEditor.Edit;
var
  C: TWPRuler;
begin
  OpenDialog := TOpenDialog.Create(nil);
  C := TWPRuler(GetComponent(0));
  with OpenDialog do
  begin
    Filter := 'Delphi project files|*.dpr|All files|*.*';
    InitialDir := C.FProjectFolder;
    Filename := 'AnyFile';
    Title := 'Locate project file';
    Options := [ofPathMustExist];
    if Execute then C.FProjectFolder := ExtractFilePath(Filename);
  end;
  OpenDialog.Free;
end;

//************ TWPSlider *******************

constructor TWPSlider.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := 9;
  Height := 8;
  FSliderStyle := ssUp;
  FFillColor := clBtnFace;
end;

destructor TWPSlider.Destroy;
begin
  inherited Destroy;
end;

procedure TWPSlider.Paint;
begin
  with Canvas do
  begin
    Brush.Color := FFillColor;
    Pen.Color := clNavy;
    case FSliderStyle of
      ssUp: begin
          MoveTo(Width shr 1, 0);
          LineTo(Width - 1, Height - 4);
          LineTo(Width - 1, Height - 1);
          LineTo(0, Height - 1);
          LineTo(0, Height - 4);
          LineTo(Width shr 1, 0);
          FloodFill(3, Height - 4, Pen.Color, fsBorder);
        end;
      ssDown: begin
          MoveTo(0, 0);
          LineTo(Width - 1, 0);
          LineTo(Width - 1, 3);
          LineTo(Width shr 1, Height - 1);
          LineTo(0, 3);
          LineTo(0, 0);
          FloodFill(3, 3, Pen.Color, fsBorder);
        end;
      ssMulti: begin
          MoveTo(0, 0);
          LineTo(Width - 1, 0);
          LineTo(Width - 1, Height - 1);
          LineTo(0, Height - 1);
          LineTo(0, 0);
          FloodFill(3, 3, Pen.Color, fsBorder);
        end;
    end;
  end;
end;

//Set the margin marker fill colour

procedure TWPSlider.SetFillColor(Value: TColor);
begin
  if Value <> FFillColor then
  begin
    FFillColor := Value;
    Invalidate;
  end;
end;

//Set the margin marker style

procedure TWPSlider.SetSliderStyle(Value: TSliderStyle);
begin
  if Value <> FSliderStyle then
  begin
    FSliderStyle := Value;
    Invalidate;
  end;
end;

//************** TWPTab *********************

constructor TWPTab.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := 6;
  Height := 6;
  FTabStyle := tsLeftAlign;
  FColor := clBlack;
  ShowHint := True;
  Hint := 'Left aligned tab';
  Left := $FFFF;
end;

destructor TWPTab.Destroy;
begin
  inherited Destroy;
end;

procedure TWPTab.Paint;
begin
  with Canvas do
  begin
    Pen.Color := FColor;
    case FTabStyle of
      tsLeftAlign: begin
          MoveTo(0, 0);
          LineTo(0, 5);
          LineTo(6, 5);
          MoveTo(1, 0);
          LineTo(1, 4);
          LineTo(6, 4);
        end;
      tsCenterAlign: begin
          MoveTo(0, 5);
          LineTo(6, 5);
          MoveTo(0, 4);
          LineTo(6, 4);
          MoveTo(2, 0);
          LineTo(2, 4);
          MoveTo(3, 0);
          LineTo(3, 4);
        end;
      tsRightAlign: begin
          MoveTo(5, 0);
          LineTo(5, 5);
          LineTo(-1, 5);
          MoveTo(4, 0);
          LineTo(4, 4);
          LineTo(-1, 4);
        end;
    end;
  end;
end;

//Set the tab colour

procedure TWPTab.SetColor(Value: TColor);
begin
  if Value <> FColor then
  begin
    FColor := Value;
    Invalidate;
  end;
end;


//Set the tab style

procedure TWPTab.SetTabStyle(Value: TTabStyle);
begin
  if Value <> FTabStyle then
  begin
    FTabStyle := Value;
    Invalidate;
  end;
end;

//***************TWPRuler ********************

constructor TWPRuler.Create(AOwner: TComponent);
var
  DC: HDC;
  N, Offset: Byte;
  TabRecs: array[0..MaxTabs - 1] of TTabRec;
  F: file of TTabRec;
  TabsLoaded: Boolean;
begin
  inherited Create(AOwner);
  DC := GetDC(0);
  FScreenRes := GetDeviceCaps(DC, LOGPIXELSX);
  ReleaseDC(0, DC);
  if not (csDesigning in ComponentState) then
    FProjectFolder := ExtractFilePath(Application.ExeName);
  AssignFile(F, FProjectFolder + DatFile);
  TabsLoaded := True;
  FNumber := 0;
  try
    Reset(F);
    for N := 0 to MaxTabs - 1 do
    begin
      Read(F, TabRecs[N]);
      if TabRecs[N].Pos > 0 then Inc(FNumber);
    end;
    CloseFile(F);
  except
    TabsLoaded := False;
  end;
  Width := 200;
  Height := 25;
  Align := alTop;
  FColor := clWindow;
  FTabColor := clBlack;
  FUnits := unCentimeters;
  FLeftInset := 10;
  FLeftMargin := 2;
  FRightMargin := 2;
  FIndent := 0;
  FNumTabs := 0;
  FShowLM := True;
  FShowRM := True;
  FShowID := True;
  FShowMU := True;
  FSort := True;
  Popup := TPopupMenu.Create(Self);
  with Popup do
    for N := 0 to 2 do
    begin
      PopupItems[N] := TMenuItem.Create(Self);
      case N of
        0: PopupItems[N].Caption := '&Left align';
{$IFDEF UKVERSION}
        1: PopupItems[N].Caption := '&Centre align';
{$ELSE}
        1: PopupItems[N].Caption := '&Center align';
{$ENDIF}
        2: PopupItems[N].Caption := '&Right align';
      end;
      PopupItems[N].OnClick := PopupClick;
      Items.Add(PopupItems[N]);
      AutoPopup := True;
    end;
  for N := 0 to MaxTabs - 1 do
  begin
    FTabObjects[N] := TWPTab.Create(Self);
    with FTabObjects[N] do
    begin
      Parent := Self;
      PopupMenu := Popup;
      OnMouseDown := TabMouseDown;
      OnMouseMove := TabMouseMove;
      OnMouseUp := TabMouseUp;
      if TabsLoaded then //Values picked up from WPRuler.dat
      begin
        TabStyle := TabRecs[N].Style;
        TabPos := TabRecs[N].Pos;
      end
      else TabStyle := tsLeftAlign;
      case TabStyle of
        tsLeftAlign: begin
            Hint := 'Left aligned tab';
            Offset := 0;
          end;
        tsCenterAlign: begin
{$IFDEF UKVERSION}
            Hint := 'Centre aligned tab';
{$ELSE}
            Hint := 'Center aligned tab';
{$ENDIF}
            Offset := 2;
          end;
      else begin
          Hint := 'Right aligned tab';
          Offset := 5;
        end;
      end;
      if TabPos = 0 then Left := $FFFF
      else if TabsLoaded then
        Left := FLeftInset - Offset + Round((TabPos * FScreenRes) / 2.54);
    end;
  end;
  with Font do
  begin
    Name := 'Arial';
    Size := 8;
    Color := clBlack;
    Style := [];
  end;
  Canvas.Font.Assign(Font);
  Dragging := False;
  TimerProc(Self);  //Get the printer page width
  FLMargin := TWPSlider.Create(Self); //Left margin marker
  with FLMargin do
  begin
    Parent := Self;
    Top := 12;
    SliderStyle := ssUp;
    Hint := 'Left margin';
    ShowHint := True;
    OnMouseDown := SliderMouseDown;
    OnMouseMove := SliderMouseMove;
    OnMouseUp := SliderMouseUp;
    Visible := FShowLM and FShowMU;
  end;
  FRMargin := TWPSlider.Create(Self); //Right margin marker
  with FRMargin do
  begin
    Parent := Self;
    Top := 12;
    SliderStyle := ssUp;
    Hint := 'Right margin';
    ShowHint := True;
    OnMouseDown := SliderMouseDown;
    OnMouseMove := SliderMouseMove;
    OnMouseUp := SliderMouseUp;
    Visible := FShowRM;
  end;
  FIMargin := TWPSlider.Create(Self); //Indent marker
  with FIMargin do
  begin
    Parent := Self;
    Top := 0;
    SliderStyle := ssDown;
    Hint := 'Left indent';
    ShowHint := True;
    OnMouseDown := SliderMouseDown;
    OnMouseMove := SliderMouseMove;
    OnMouseUp := SliderMouseUp;
    Visible := FShowID;
  end;
  FMulti := TWPSlider.Create(Self); //Left margin plus indent marker
  with FMulti do
  begin
    Parent := Self;
    Top := FLMargin.Top + FLMargin.Height - 1;
    Height := 6;
    SliderStyle := ssMulti;
    Hint := 'Left margin + indent';
    ShowHint := True;
    OnMouseDown := SliderMouseDown;
    OnMouseMove := SliderMouseMove;
    OnMouseUp := SliderMouseUp;
    Visible := FShowMU;
  end;
  Timer := TTimer.Create(Self);
  with Timer do
  begin
    Interval := 500;
    OnTimer := TimerProc;
    Enabled := True;
  end;
end;

destructor TWPRuler.Destroy;
var
  N: Byte;
begin
  Timer.Free;
  FLMargin.Free;
  FRMargin.Free;
  FIMargin.Free;
  Popup.Free;
  for N := 0 to MaxTabs - 1 do FTabObjects[N].Free;
  inherited Destroy;
end;

//Tidy up after loading

procedure TWPRuler.Loaded;
begin
  inherited Loaded;
  FNumTabs := FNumber;
  if FProjectFolder = '' then
    ProjectFolder := ExtractFilePath(Application.ExeName);
end;

procedure TWPRuler.Paint;
var
  Mult: Single;
  N, X, XMark, XPos, XDot, Y, YMark, YDot, Num, Wid, Len, Ht, DotHt, Bot: Word;
  S: string;
  DC: HDC;
begin
  Canvas.Font.Assign(Font);
  DC := Canvas.Handle;
  Bot := Height - 7;
  Ht := Canvas.TextHeight('1');
  Y := ((Bot - 2 - Ht) shr 1) + 2;
  DotHt := Ht div 5;  //Height of the dot marks
  Ht := Ht div 3;     //Height of the halfway marks
  YMark := ((Bot - 2 - Ht) shr 1) + 2;
  YDot := ((Bot - 2 - DotHt) shr 1) + 2;
  if FUnits = unInches then Mult := 1 else Mult := 1 / 2.54;
  with Canvas do  //Draw the ruler outline
  begin
    Pen.Color := clBtnShadow;
    Rectangle(FLeftInset - 1, 1, Width, Bot + 1);
    Pen.Color := clBlack;
    Brush.Color := FColor;
    Rectangle(FLeftInset, 2, Width - 1, Bot);
    Pen.Color := clBtnHighlight;
    MoveTo(FLeftInset + 1, Bot - 1);
    LineTo(Width - 2, Bot - 1);
    LineTo(Width - 2, 2);
    Pen.Color := clBtnHighlight;
    Brush.Color := clBtnFace;
    Wid := FLeftInset + Round(FPageWidth * FScreenRes * Mult); //End of printable area
    Rectangle(Wid - 1, 2, Width - 1, Bot);
  end;
  Num := 0;
  Wid := FScreenRes shr 1; //Number of pixels in half an inch
  SelectObject(DC, CreatePen(PS_SOLID, 1, Canvas.Font.Color));
  SetBkMode(DC, TRANSPARENT);
  repeat  //Draw the ruler markings
    Inc(Num);
    X := Num * FScreenRes; //X offset of center of number mark (inch basis)
    XMark := X - Wid; //Ditto for half-inch mark
    if FUnits = unCentimeters then
    begin
      X := Round(X / 2.54);
      XMark := Round(XMark / 2.54);
    end;
    XMark := XMark + FLeftInset; //X position of half-inch or half-cm mark
    S := IntToStr(Num);
    Len := Canvas.TextWidth(S);
    XPos := X + LeftInset; //X Position of center of number mark
    X := XPos - (Len shr 1) + 1; //X position of number mark
    TextOut(DC, X, Y, PChar(S), Length(S));
    MoveToEx(DC, XMark, YMark, nil);
    LineTo(DC, XMark, YMark + Ht);
    if FUnits = unInches then
      for N := 1 to 3 do //Draw 1/8 inch marks
      begin
        Len := N * (Wid shr 2);
        XDot := XMark + Len;
        MoveToEx(DC, XDot, YDot, nil);
        LineTo(DC, XDot, YDot + DotHt);
        XDot := XMark - Len;
        MoveToEx(DC, XDot, YDot, nil);
        LineTo(DC, XDot, YDot + DotHt);
      end
    else begin //Draw 1/4 cm marks
      XDot := FLeftInset + Round((Num - 0.75) * FScreenRes * Mult);
      MoveToEx(DC, XDot, YDot, nil);
      LineTo(DC, XDot, YDot + DotHt);
      XDot := FLeftInset + Round((Num - 0.25) * FScreenRes * Mult);
      MoveToEx(DC, XDot, YDot, nil);
      LineTo(DC, XDot, YDot + DotHt);
    end;
  until X > Width;
  DeleteObject(GetStockObject(BLACK_PEN));
  if Dragging then Exit;
  FLMargin.Top := Bot - 6;
  FRMargin.Top := FLMargin.Top;
  FMulti.Top := FLMargin.Top + FLMargin.Height - 1;
  FLMargin.Left := FLeftInset + Round(FLeftMargin * Mult * FScreenRes)
    - (FLMargin.Width shr 1);
  FMulti.Left := FLMargin.Left;
  FIMargin.Left := FLeftInset - (FLMargin.Width shr 1)
    + Round((FLeftMargin + FIndent) * Mult * FScreenRes);
  FRMargin.Left := FLeftInset + Round((FPageWidth - FRightMargin)
    * Mult * FScreenRes) - (FRMargin.Width shr 1);
  for N := 0 to MaxTabs - 1 do FTabObjects[N].Top := FLMargin.Top;
end;

//Refresh the printer page width

procedure TWPRuler.TimerProc(Sender: TObject);
var
  Wid: Single;
begin
  if Printer.Printers.Count = 0 then Wid := 20
  else Wid := GetDeviceCaps(Printer.Handle, HORZSIZE) / 10;
  if Wid <> PrinterWidth then
  begin
    PrinterWidth := Wid;
    if FUnits = unCentimeters then FPageWidth := PrinterWidth
    else FPageWidth := PrinterWidth / 2.54;
    Invalidate;
  end;
end;

//Change tab style from popup menu

procedure TWPRuler.PopupClick(Sender: TObject);
var
  N: Byte;
  Edge: Word;
begin
  for N := 0 to 2 do
    if Sender = PopupItems[N] then
      with Popup.PopupComponent as TWPTab do
      begin
        case TabStyle of
          tsLeftAlign: Edge := Left;
          tsCenterAlign: Edge := Left + 2;
        else Edge := Left + 5;
        end;
        TabStyle := TTabStyle(N);
        case N of
          0: begin
              Hint := 'Left aligned tab';
              Left := Edge;
            end;
          1: begin
{$IFDEF UKVERSION}
              Hint := 'Centre aligned tab';
{$ELSE}
              Hint := 'Center aligned tab';
{$ENDIF}
              Left := Edge - 2;
            end;
          2: begin
              Hint := 'Right aligned tab';
              Left := Edge - 5;
            end;
        end;
        Break;
      end;
end;

//Add a tab with a program command

function TWPRuler.AddTab(Position: Single; Style: TTabStyle): Boolean;
begin
  if NumTabs = MaxTabs then Result := False
  else begin
    Result := True;
    AddOneTab(Position, Style);
  end;
end;

//Add a tab with a user click

procedure TWPRuler.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  Mult: Single;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if FNumTabs = MaxTabs then Exit;
  if FUnits = unInches then Mult := 1 else Mult := 2.54;
  if Button = mbLeft then
    AddOneTab((X - 1 - FLeftInset) * Mult / FScreenRes, tsLeftAlign);
end;

procedure TWPRuler.AddOneTab(X: Single; Style: TTabStyle);
var
  Mult: Single;
  N, Offset: Byte;
begin
  case Style of
    tsLeftAlign: Offset := 0;
    tsCenterAlign: Offset := 2;
  else Offset := 5;
  end;
  with FTabObjects[FNumTabs] do
  begin
    if FUnits = unInches then Mult := 1 else Mult := 1 / 2.54;
    Left := FLeftInset - Offset + Round(X * FScreenRes * Mult);
    TabStyle := Style;
    TabStyles[FNumTabs] := Style;
    TabValues[FNumTabs] := X;
    Inc(FNumTabs);
    SortTabs;
    N := 0;
    while TabPos < X do Inc(N);
    if Assigned(FOnTabAdded) then FOnTabAdded(Self as TObject, N);
  end;
end;

//Remove a tab with a program command

function TWPRuler.RemoveTab(Index: Byte): Boolean;
begin
  if (Index > MaxTabs - 1) or (FTabObjects[Index].TabPos = 0) then
  begin
    Result := False;
    Exit;
  end;
  RemoveOneTab(Index);
  Result := True;
end;

//Remove a tab

procedure TWPRuler.RemoveOneTab(Index: Byte);
var
  N: Byte;
  Tab: TWPTab;
begin
  if Index < MaxTabs - 1 then
  begin
    Tab := FTabObjects[Index];
    for N := Index to MaxTabs - 2 do FTabObjects[N] := FTabObjects[N + 1];
    FTabObjects[MaxTabs - 1] := Tab;
  end;
  FTabObjects[MaxTabs - 1].Left := $FFFF;
  FTabObjects[MaxTabs - 1].TabPos := 0;
  FTabObjects[MaxTabs - 1].TabStyle := tsLeftAlign;
  Dec(FNumTabs);
end;

//Clear all tabs with a program command

procedure TWPRuler.ClearAllTabs;
var
  N: Byte;
begin
  for N := 0 to MaxTabs - 1 do
  begin
    FTabObjects[N].Left := $FFFF;
    FTabObjects[N].TabPos := 0;
    FTabObjects[N].TabStyle := tsLeftAlign;
  end;
end;

//Start to drag a margin marker

procedure TWPRuler.SliderMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Dragging then Exit;
  Dragging := True;
  Screen.Cursor := crDrag;
  Diff := FIMargin.Left - FLMargin.Left;
end;

//Drag a margin marker

procedure TWPRuler.SliderMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  XPos: Integer;
begin
  if Dragging then
  begin
    Screen.Cursor := crDrag;
    with Sender as TWPSlider do
    begin
      XPos := X + Left;
      if ((Sender = FLMargin) or (Sender = FIMargin) or (Sender = FRMargin))
        and (XPos < FLeftInset)
        then TWPSlider(Sender).Left := FLeftInset
        - (TWPSlider(Sender).Width shr 1)
      else if (Sender = FMulti)
        and ((XPos < FLeftInset)
        or (XPos + Diff < FLeftInset))
        then begin
        if FMulti.Left <= FIMargin.Left then
          FMulti.Left := FLeftInset - (FMulti.Width shr 1)
        else FMulti.Left := FLeftInset - Diff - (FMulti.Width shr 1);
      end
      else Left := XPos - (Width shr 1);
    end;
    if Sender = FLMargin then FMulti.Left := FLMargin.Left;
    if Sender = FMulti then
    begin
      FIMargin.Left := FMulti.Left + Diff;
      FLMargin.Left := FMulti.Left;
    end;
  end;
end;

//Complete the dragging of a margin marker

procedure TWPRuler.SliderMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  XPos: Word;
  Mult: Single;
begin
  with Sender as TWPSlider do XPos := Left + (Width shr 1) - FLeftInset;
  if FUnits = unInches then Mult := 1 else Mult := 2.54;
  if Sender = FLMargin then
  begin
    FIndent := FIndent + FLeftMargin - (XPos * Mult / FScreenRes);
    FLeftMargin := XPos * Mult / FScreenRes;
    if Assigned(FOnSliderMoved) then
    begin
      FOnSliderMoved(Self as TObject, 'LM', FLeftMargin);
      FOnSliderMoved(Self as TObject, 'ID', FIndent);
    end;
  end
  else if Sender = FMulti then
  begin
    FLeftMargin := XPos * Mult / FScreenRes;
    if Assigned(FOnSliderMoved) then
      FOnSliderMoved(Self as TObject, 'LM', FLeftMargin);
  end
  else if Sender = FRMargin then
  begin
    FRightMargin := FPageWidth - (XPos * Mult / FScreenRes);
    if Assigned(FOnSliderMoved) then
      FOnSliderMoved(Self as TObject, 'RM', FRightMargin);
  end
  else begin
    FIndent := (XPos * Mult / FScreenRes) - FLeftMargin;
    if Assigned(FOnSliderMoved) then
      FOnSliderMoved(Self as TObject, 'ID', FIndent);
  end;
  Dragging := False;
  Invalidate;
  Screen.Cursor := crDefault;
end;

//Start to drag a tab

procedure TWPRuler.TabMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Dragging then Exit;
  Dragging := True;
  Screen.Cursor := crDrag;
end;

//Drag a tab

procedure TWPRuler.TabMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  XPos: Integer;
begin
  if Dragging then
  begin
    Screen.Cursor := crDrag;
    with Sender as TWPTab do
    begin
      if (Y < -Top) or (Y > TWPRuler(Parent).Height - Top) then Color := clGray
      else Color := FTabColor;
      XPos := X + Left;
      if XPos < FLeftInset then XPos := FLeftInset;
      Left := XPos - (Width shr 1);
    end;
  end;
end;

//Complete the dragging of a tab

procedure TWPRuler.TabMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  XPos: Word;
  Mult: Single;
  N: Byte;
begin
  with Sender as TWPTab do XPos := Left + (Width shr 1) - FLeftInset;
  if FUnits = unInches then Mult := 1 else Mult := 2.54;
  for N := 0 to MaxTabs - 1 do
    if FTabObjects[N] = Sender then
    begin
      with FTabObjects[N] do
        if (Y < -Top) or (Y > TWPRuler(Parent).Height - Top) then RemoveOneTab(N)
        else begin
          TabPos := XPos * Mult / FScreenRes;
          SortTabs;
        end;
      if Assigned(FOnTabMoved) then FOnTabMoved(Self as TObject, N);
      Break;
    end;
  Dragging := False;
  Invalidate;
  Screen.Cursor := crDefault;
end;

//Sort the tabs into ascending order

procedure TWPRuler.SortTabs;
var
  Tab: TWPTab;
  N: Byte;
  Pos0, Pos1: Single;
  Changed: Boolean;
begin
  if not FSort then Exit;
  repeat
    Changed := False;
    for N := 0 to MaxTabs - 2 do
    begin
      Pos0 := FTabObjects[N].TabPos;
      if Pos0 = 0 then Pos0 := 999999;
      Pos1 := FTabObjects[N + 1].TabPos;
      if Pos1 = 0 then Pos1 := 999999;
      if Pos0 > Pos1 then
      begin
        Changed := True;
        Tab := FTabObjects[N];
        FTabObjects[N] := FTabObjects[N + 1];
        FTabObjects[N + 1] := Tab;
      end;
    end;
  until not Changed;
end;

//Dummy read/write routines for property Tabs

procedure TWPRuler.SetTabs(Value: TDummy);
begin
end;

function TWPRuler.GetTabs: TDummy;
begin
  Result := nil;
end;

//Count the tabs and save the settings to DatFile

procedure TWPRuler.DoTabs(Value: Boolean);
var
  N: Byte;
  F: file of TTabRec;
  Rec: TTabRec;
begin
  SortTabs;
  FNumTabs := 0;
  while (FNumTabs < MaxTabs) and (FTabObjects[FNumTabs].TabPos > 0) do Inc(FNumTabs);
  if FNumTabs < MaxTabs then
    for N := FNumTabs to MaxTabs - 1 do FTabObjects[N].Left := $FFFF;
  if FProjectFolder = '' then
  begin
    ShowMessage('ProjectFolder property has not been defined. Unable to save tabs.');
    Exit;
  end;
  if FProjectFolder[Length(FProjectFolder)] <> '\' then
    FProjectFolder := FProjectFolder + '\';
  AssignFile(F, FProjectFolder + DatFile);
  try
    Rewrite(F);
    try
      for N := 0 to MaxTabs - 1 do
      begin
        with Rec do
        begin
          Pos := FTabObjects[N].TabPos;
          Style := FTabObjects[N].TabStyle;
        end;
        Write(F, Rec);
      end;
    finally
      CloseFile(F);
    end;
  except
    ShowMessage('Unable to save tab values. ProjectFolder does not contain a valid path.');
  end;
end;

//Get the screen pixel position of a tab

function TWPRuler.GetTabPixels(Index: Byte): Word;
begin
  Result := FTabObjects[Index].Left;
end;

//Get the style of a tab

function TWPRuler.GetTabStyles(Index: Byte): TTabStyle;
begin
  Result := FTabObjects[Index].TabStyle;
end;

//Get the value of a tab

function TWPRuler.GetTabValues(Index: Byte): Single;
begin
  Result := FTabObjects[Index].TabPos;
end;

//Set the background colour

procedure TWPRuler.SetColor(Value: TColor);
begin
  if Value <> FColor then
  begin
    FColor := Value;
    Invalidate;
  end;
end;

//Set the Sort property
procedure TWPRuler.SetSort(Value: Boolean);
begin
  if Value <> FSort then
  begin
    FSort := Value;
    if Value then SortTabs;
  end;
end;

//Set the tab colour
procedure TWPRuler.SetTabColor(Value: TColor);
var
  N: Byte;
begin
  if Value <> FTabColor then
  begin
    FTabColor := Value;
    for N := 0 to MaxTabs - 1 do FTabObjects[N].Color := Value;
  end;
end;

//Set the style of a tab

procedure TWPRuler.SetTabStyles(Index: Byte; Value: TTabStyle);
var
  OldOffset: Byte;
begin
  if Value <> FTabObjects[Index].TabStyle then
  begin
    case FTabObjects[Index].TabStyle of
      tsLeftAlign: OldOffset := 0;
      tsCenterAlign: OldOffset := 2;
    else OldOffset := 5;
    end;
    with FTabObjects[Index] do
    begin
      TabStyle := Value;
      case Value of
        tsLeftAlign: begin
            Hint := 'LeftAligned tab';
            Left := Left + OldOffset;
          end;
        tsCenterAlign: begin
{$IFDEF UKVERSION}
            Hint := 'Centre aligned tab';
{$ELSE}
            Hint := 'Center aligned tab';
{$ENDIF}
            Left := Left + OldOffset - 2;
          end;
      else begin
          Hint := 'Right aligned tab';
          Left := Left + OldOffset - 5;
        end;
      end;
      Invalidate;
    end;
  end;
end;

//Change a tab value

procedure TWPRuler.SetTabValues(Index: Byte; Value: Single);
var
  Mult: Single;
  Offset: Byte;
begin
  if Value <> FTabObjects[Index].TabPos then
  begin
    FTabObjects[Index].TabPos := Value;
    if FUnits = unInches then Mult := 1 else Mult := 1 / 2.54;
    with FTabObjects[Index] do
    begin
      if TabStyle = tsLeftAlign then Offset := 0
      else if TabStyle = tsCenterAlign then Offset := 2
      else Offset := 5;
      if Value = 0 then Left := $FFFF
      else Left := FLeftInset - Offset + Round(Value * FScreenRes * Mult);
    end;
    SortTabs;
  end;
end;

//Set the indent value

procedure TWPRuler.SetIndent(Value: Single);
begin
  if Value = FIndent then Exit;
  if FLeftMargin + Value < 0 then Exit;
  if FLeftMargin + Value > FPageWidth then Exit;
  FIndent := Value;
  Invalidate;
end;

//Set the left margin value

procedure TWPRuler.SetLeftMargin(Value: Single);
begin
  if Value = FLeftMargin then Exit;
  if Value > FPageWidth then Exit;
  if Value >= FPageWidth - FRightMargin then Exit;
  if Value + FIndent > FPageWidth then Exit;
  if Value < 0 then Exit;
  if Value + FIndent < 0 then Exit;
  FLeftMargin := Value;
  Invalidate;
end;

//Set the left inset value

procedure TWPRuler.SetLeftInset(Value: Word);
begin
  if Value <> FLeftInset then
  begin
    FLeftInset := Value;
    Invalidate;
  end;
end;

//Set the right margin value

procedure TWPRuler.SetRightMargin(Value: Single);
begin
  if Value = FRightMargin then Exit;
  if Value < 0 then Exit;
  if FPageWidth - Value <= FLeftMargin then Exit;
  FRightMargin := Value;
  Invalidate;
end;

//Show or hide the left margin marker

procedure TWPRuler.SetShowLM(Value: Boolean);
begin
  if Value <> FShowLM then
  begin
    FShowLM := Value;
    FLMargin.Visible := Value;
    if not Value then
    begin
      FMulti.Visible := False;
      FShowMU := False;
    end;
    Invalidate;
  end;
end;

//Show or hide the right margin marker

procedure TWPRuler.SetShowRM(Value: Boolean);
begin
  if Value <> FShowRM then
  begin
    FShowRM := Value;
    FRMargin.Visible := Value;
    Invalidate;
  end;
end;

//Show or hide the indent marker

procedure TWPRuler.SetShowID(Value: Boolean);
begin
  if Value <> FShowID then
  begin
    FShowID := Value;
    FIMargin.Visible := Value;
    Invalidate;
  end;
end;

//Show or hide the "Left margin plus indent" marker

procedure TWPRuler.SetShowMU(Value: Boolean);
begin
  if Value <> FShowMU then
  begin
    FShowMU := Value;
    FMulti.Visible := Value;
    if Value then FLMargin.Visible := True;
    Invalidate;
  end;
end;

//Set the measurement units

procedure TWPRuler.SetUnits(Value: TUnits);
var
  Mult: Single;
  N: Byte;
begin
  if Value <> FUnits then
  begin
    FUnits := Value;
    if Value = unInches then Mult := 1 / 2.54
    else Mult := 2.54;
    FLeftMargin := FLeftMargin * Mult;
    FRightMargin := FRightMargin * Mult;
    FIndent := FIndent * Mult;
    FPageWidth := FPageWidth * Mult;
    for N := 0 to MaxTabs - 1 do
      FTabObjects[N].TabPos := FTabObjects[N].TabPos * Mult;
    Invalidate;
  end;
end;

end.

